#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# BrEPSdb creates the enzyme sequence pattern database BrEPS.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``brepsdb.py``
------------
This is the main file to call for BrEPS database creation.

For desktop mode just call ``python breps.py`` in the command line,
for the computer cluster (SGE) mode should be deployed via ``qsub``
from cmaster01 using ``sge_starter.bash breps.py``. For example: ``qsub
sge_starter.bash breps.py``

No further functions or classes are provided.
"""
import os
import sys
import json
import logging
import datetime
import MySQLdb
import logging.config

import data
import blast
import clustering
import pattern
import verification
import createdb

import brepslib as bl

version = "2.0.0"

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

if __name__ == "__main__":

    print(u"BrEPSdb Copyright (C) 2017 Christian-Alexander Dudek,\n"
          u"Technische Universität Braunschweig\n"
          u"Department of Bioinformatics and Biochemistry\n\n"
          u"This program comes with ABSOLUTELY NO WARRANTY!\n"
          u"See https://opensource.org/licenses/GPL-3.0 for detailed "
          u"license information.\n")

    lstSteps = [0, 1, 2, 3, 4, 5]
    blnUseSGE = dctCfg["use_sge"]
    blnAdvanced = dctCfg["advanced"]

    # We want to check the installation before starting BrEPS.
    # Create all folders which do not exist.
    log.info("Checking folders...")
    lstFolders = ["data", "pattern", "export", "logs"]
    for f in lstFolders:
        if not os.path.isdir(f):
            os.makedirs(f)
            log.info("Folder {} created!".format(f))

    # BLAST and ClustalO executables are needed/linked in the
    # corresponding directories.
    log.info("Checking external executbales...")
    lstExecutables = [
        "blast-current/bin/blastp",
        "blast-current/bin/makeblastdb",
        "blast-current/bin/segmasker",
        "clustal-current/bin/clustalo"
        ]
    blnStop = False
    for e in lstExecutables:
        if not os.path.exists(e):
            blnStop = True
            log.error("Executable {0} not found! Please "
                      "link {0} to {1}!".format(
                            e.split("/")[-1], e.replace(e.split("/")[-1], "")))

    # Check if we can connect to the database!
    log.info("Checking database access...")
    try:
        cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])
    except MySQLdb.OperationalError as e:
        log.error(e[1])
        blnStop = True
    else:
        cur.close()

    if blnStop:
        log.error("Requirements not met. Please check "
                  "errors above and start again!")
        exit()

    log.info("Starting BrEPSdb...")
    log.info("SGE parallelization: {}".format(blnUseSGE))
    log.info("Advanced data preparation: {}".format(blnAdvanced))

    # Step 1: Data download and parsing
    if 0 in lstSteps:
        if blnUseSGE:
            # We need to send this step to the cmaster01, because file
            # downloads and copying of files can only be done via
            # cmaster01.
            os.system(
                "ssh cmaster01 \"cd {0}; python {0}/data.py\"".format(strPath)
                )
        else:
            # For desktop mode we just call data.py directly, to avoid
            # redundant sourcecode for the same tasks!
            os.system("cd {0}; python {0}/data.py".format(strPath))
        # Before we can proceed to the next step, we want to check if
        # the data preparation was successful.
        if not os.path.isfile("{}/data/done.tmp".format(strPath)):
            log.error("An error occured in data preparation...")
            sys.exit()
        else:
            os.system("rm {}/data/done.tmp".format(strPath))
    # Step 2: BLAST
    if 1 in lstSteps:
        lstFiles = blast.prepare(100)
        lstFiles = blast.blast(lstFiles, useSGE=blnUseSGE)
        blast.save(lstFiles)

    # Step 3: Clustering
    if 2 in lstSteps:
        dctClusters = clustering.agglomerate(threshold=0.0001)
        clustering.save(dctClusters)

    # Step 4: Pattern generation
    if 3 in lstSteps:
        pattern.generate(4, useSGE=blnUseSGE)
        pattern.extend()

    # Step 5: Verification
    if 4 in lstSteps:
        verification.verificator("patterns", 4, useSGE=blnUseSGE)
        verification.verificator("patterns_extended", 4, useSGE=blnUseSGE)

    # Step 6: Database creation
    if 5 in lstSteps:
        # Prepare release data
        cur = bl.Dbase("mysql", dctCfg["final_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])
        now = datetime.datetime.now()
        strYear = now.strftime("%Y")
        strDate = now.strftime("%Y-%m-%d")
        cur.execute("SELECT serial FROM releaseinfo WHERE year=%s "
                    "ORDER BY rel_id DESC LIMIT 1", (strYear,))
        dctRes = cur.fetchone()
        if dctRes:
            intSerial = dctRes[0]+1
        else:
            intSerial = 1
        cur.close()
        strUniDate, strUni = data.get_release(
            "{}/data/reldate.txt".format(strPath)
            )
        # TODO: add statistics, adding a chart?
        lstRel = [int(strYear), intSerial, version, strDate, strUni,
                  0, 0, 0, 0, 0, "original"]

        # Prepare databases
        createdb.prepare_db()
        lstAnnotation, lstPatterns = createdb.prepare_data()

        createdb.save("patterns", lstPatterns, createdb.lstPatFields)
        createdb.save("annotation", lstAnnotation, createdb.lstFields)
        createdb.save("releaseinfo", (lstRel,), createdb.lstRelFields)

        # Export MySQL file
        os.system("mysqldump -h {0} -u {1} -p{2} {3} > {4}/{5}/{3}.sql".format(
            dctCfg["db_host"], dctCfg["db_user"], dctCfg["db_pass"],
            dctCfg["final_db"], strPath, dctCfg["export_dir"]))
        os.system("mv {0}.db {1}/{2}/{0}.db".format(
            dctCfg["final_db"], strPath, dctCfg["export_dir"]))
        os.system("gzip {}/{}/{}.sql".format(
            strPath, dctCfg["export_dir"], dctCfg["final_db"]
            ))
        os.system("gzip {}/{}/{}.db".format(
            strPath, dctCfg["export_dir"], dctCfg["final_db"]
            ))
     # TODO: cleanup

    log.info("BrEPS database susccessfully created!")
