#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``brepslib.py``
-----------------
This library contains regular expressions, functions and classes used by
multiple BrEPS modules.
"""
import os
import re
import sys
import sqlite3 as sqlt
import MySQLdb as mysql

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

reRep = re.compile("\((|\d+,)?(\d+)\)-?")
reSel = re.compile("-?\[([ARNDCQEGHILKMFPSTWYVBZX]*)\]-?")
reAa = re.compile("-?([ARNDCQEGHILKMFPSTWYVBZX])-?")
reAny = re.compile("-?([x])-?")


class Dbase(object):
    """
    Database class for both MySQL and SQLite usage.

    On creation the ``strType`` can be selected from ``MySQL`` and
    ``SQLite`` to select a database type. All functions are created just
    like the provided by MySQLdb or sqlite3. MySQL placeholders ``%s``
    are automatically replaced for SQLite. Before every execution the
    connection is checked to avoid ``mysql.OperationalError``
    exceptions. The returned instance can be directly used for
    execution. Also the instance needs to be closed properly.

    **Only non-standard methods are documented!**

    Attributes:
        strType (str): Type of database connection (mysql or sqlite).
        strDB (str): Database name for MySQL, path to file for SQLite.
        strHost (str): Host for the MySQL connection.
        strUser (str): Username for the MySQL connection.
        strPass (str): Password for the MySQL connection.

    """
    def __init__(self, strType, strDB, strHost=None,
                 strUser=None, strPass=None):
        super(Dbase, self).__init__()
        self.strType = strType
        self.strDB = strDB
        self.strHost = strHost
        self.strUser = strUser
        self.strPass = strPass
        self.connect()
        self.description = None

    def connect(self):
        if self.strType == "mysql":
            self.db = mysql.connect(self.strHost,
                                    self.strUser,
                                    self.strPass,
                                    self.strDB)
            self.s = "%s"
        elif self.strType == "sqlite":
            self.db = sqlt.connect(self.strDB)
            self.s = "?"
        self.cur = self.db.cursor()

    def execute(self, strQry, lstData=[], commit=True):
        try:
            res = self.cur.execute(strQry.replace("%s", self.s), lstData)
        except mysql.OperationalError as e:
            if e[0] == 2006:
                self.db.close()
                self.connect()
                res = self.execute(strQry.replace("%s", self.s), lstData)
            else:
                raise e
        self.description = self.cur.description
        if commit:
            self.db.commit()
        return res

    def executemany(self, strQry, lstData=[], commit=True):
        # We want to split huge queries into smaller ones...
        # Large queries can cause Error 2006, (MySQl server has gone
        # away)!

        if lstData:
            lstChunks = [lstData[i:i+200]
                         for i in xrange(0, len(lstData), 200)]
            for chunk in lstChunks:
                try:
                    res = self.cur.executemany(strQry.replace("%s", self.s),
                                               chunk)
                except mysql.OperationalError as e:
                    if e[0] == 2006:
                        self.connect()
                        res = self.cur.executemany(strQry.replace("%s",
                                                   self.s), chunk)
                    else:
                        raise e
                self.description = self.cur.description
            if commit:
                self.commit()

            return res
        else:
            return None

    def load_sql(self, strFile):
        os.system(
            "mysql -h %s -u %s, -p%s %s < %s".format(
                self.strHost, self.strUser, self.strPass, self.strDB, strFile
                )
            )

    def fetchall(self):
        return self.cur.fetchall()

    def fetchone(self):
        return self.cur.fetchone()

    def commit(self):
        self.db.commit()

    def close(self):
        self.cur.close()
        self.db.close()


def subpattern(strPS):
    """
    Creates a low complexity subpattern for pattern pre checks.

    Args:
        strPS (str): Pattern in PROSITE pattern format.

    Returns:
        str: Subpattern of ``strPS`` in PROSITE format.

    """
    lstPS = strPS.split("-")

    lstPos = []
    lstSubs = []

    for item in lstPS:
        if "," in item:
            lstSubs.append(lstPos)
            lstPos = []
        else:
            lstPos.append(item)

    # if we have no variable x positions, we simply split the pattern
    if not lstSubs:
        lstSubs = [lstPS[i:i+10] for i in xrange(0, len(lstPS), 10)]
    maxPos = min(10, max([len(i) for i in lstSubs]))

    minComp = 1
    bestSub = ""
    for item in lstSubs:
        if len(item) >= maxPos:
            curStr = "-".join(item)
            curComp = complexity(curStr)
            if curComp < minComp:
                bestSub = curStr
                minComp = curComp
    return bestSub


def ps2regex(strProsite):
    """
    Converts PROSITE patterns to regular expressions.

    Args:
        strProsite (str): The PROSITE pattern.

    Returns:
        str: Regular expression representation of the pattern.

    """
    strRegex = reSel.sub("[\\1]", strProsite)
    strRegex = reRep.sub("{\\1\\2}", strRegex)
    strRegex = reAa.sub("\\1", strRegex)
    strRegex = reAny.sub(".", strRegex)

    return strRegex


def ps2consensus(strProsite):
    """
    Converts PROSITE patterns to consensus representation.

    Args:
        strProsite (str): The PROSITE pattern.

    Returns:
        str: Consensus representation of the pattern.

    """
    strRegex = reSel.sub("[\\1]", strProsite)
    strRegex = reRep.sub("{\\1\\2}", strRegex)
    strRegex = reAa.sub("\\1", strRegex)
    strRegex = reAny.sub(".", strRegex)

    return strRegex


def complexity(strPS):
    """
    Calculates the complexity from a PROSITE pattern.

    The pattern complexity is calculated for every amino acid by the
    following rules:

        - A conserved amino acid has complexity 0.0.
        - Set of n aminoacids has complexity of n/20.
        - Wildcard positions (x) have complexity of 1.0.

    The sum of all complexities will be divided by the maximum length
    of the pattern.

    Args:
        strPS (str): Patten in PROSITE format.

    Returns:
        float: Complexity ranging from 1.0 (complex) to 0.0 (not
            complex).
    """
    lstPS = strPS.split("-")
    complexity = 0.0
    min_len = 0
    max_len = 0
    for pos in lstPS:
        if "[" in pos:
            complexity += (len(pos)-2)/20.0
            min_len += 1
            max_len += 1
        elif "x" in pos and "," in pos:
            n = pos.replace("x(", "").replace(")", "").split(",")
            mn = int(n[0])
            mx = int(n[1])
            complexity += mx*1.0
            min_len += mn
            max_len += mx
        elif "x" in pos and "(" in pos:
            num = int(pos.replace("x(", "").replace(")", ""))
            complexity += num*1.0
            min_len += num
            max_len += num
        elif "x" == pos:
            complexity += 1.0
            min_len += 1
            max_len += 1
        else:
            complexity += 1.0/20
            min_len += 1
            max_len += 1
    # Normalize the complexity by the maximum length
    return complexity/max_len

if __name__ == "__main__":
    pass
