#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``clustering.py``
-----------------
This module clusters the sequences using the BLAST aligbnment data.
The module implements the optimally efficient CLINK algorithm using the
MySQL database as data backend. The result is a database assisted CLINK
algorithm, specialized for the usage of BLAST E-values as distance
measurement.

The ``data`` folder is associated to this module.
"""
import os
import sys
import json
import logging
import logging.config

import brepslib as bl

inf = float("infinity")
COMPLETE_LINKAGE = 0
SINGLE_LINKAGE = 1
AVERAGE_LINKAGE = 2

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class Node(object):
    """Represents a node in a cluster.

    The outer node is the created cluster. Every node has a left and
    right child node. If not the node is a leaf in the tree.

    Attributes:
        seq_id (int): Sequence id, if leaf node, ``None`` otherwise.
        left (int): Left child node, or ``None`` if node is a leaf.
        right (int): Right child node, or ``None`` if node is a leaf.
        dist (float): Distance between left and right child node
            (E-value).
        depth (int): Maximum num of edges downstream of the node.
        size (int): Num of nodes downstream of the node.
    """

    def __init__(self, seq_id, left, right, dist, node_id):
        """Inits the node and calculates depth and nodes and leafs."""
        super(Node, self).__init__()
        self.node_id = node_id
        self.seq_id = seq_id
        self.left = left
        self.right = right
        self.dist = dist
        self.parent_id = None
        if self.is_leaf():
            self.depth = 0
            self.leafs = 1
            self.nodes = 1
        else:
            self.nodes = left.nodes + right.nodes + 1
            self.leafs = left.leafs + right.leafs
            self.depth = 1 + max(left.depth, right.depth)

    def __len__(self):
        """Returns the number of nodes beneath this node."""
        return self.nodes - 1

    def __str__(self):
        """Returns a string with nodes, leafs and depth of the node."""
        return "Node {} with {} nodes, {} leafs and depth {}".format(
            self.node_id, self.nodes, self.leafs, self.depth)

    def get_nodes(self, lstNodes=None):
        """Returns all nodes as a list."""
        if lstNodes is None:
            lstNodes = []

        if self.is_leaf():
            lstNodes.append(self)
            return lstNodes
        else:
            lstNodes.append(self)
            lstNodes = self.left.get_nodes(lstNodes)
            lstNodes = self.right.get_nodes(lstNodes)
        return lstNodes

    def get_leafs(self):
        return [x for x in self.get_nodes() if x.is_leaf()]

    def to_list(self):
        """
        Returns a nested list of all leaf seq_ids.

        A list represents a couple of nodes, for example:
        [[[[[0, 1], [2, 3]], [6, 7]], [4, 5]], [8, 9]]
        """
        if self.is_leaf():
            return self.seq_id
        else:
            return [self.left.to_list(), self.right.to_list()]

    def is_leaf(self):
        """Returns True if the node has no child nodes. Otherwise False."""
        if self.left is None and self.right is None:
            return True
        else:
            return False

    def to_tuple(self, cluster_id):
        """
        Returns a tuple with values of all attributes.

        Needs the ``cluster_id`` for proper include into the database!
        """
        if self.is_leaf():
            return (self.node_id, cluster_id, self.parent_id, None, None,
                    self.seq_id, self.dist, self.depth, len(self))
        else:
            return (self.node_id, cluster_id, self.parent_id,
                    self.left.node_id, self.right.node_id, self.seq_id,
                    self.dist, self.depth, len(self))


def agglomerate(threshold=0.0001):
    """
    Agglomerates the given data into one or more clusters.

    Uses clink clustering as described in
    http://comjnl.oxfordjournals.org/content/16/1/30.short
    and http://comjnl.oxfordjournals.org/cgi/content/long/20/4/364

    Args:
        threshold (float): Defines the threshold where the
            agglomeration. will stop. Resulting in disconnected trees.
            (Default: 1e-4).

    Returns:
        dict: Returns a dictionary with all created clusters.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    # We copy the table alignments to distances, as we need to change
    # the evalues while clustering.

    log.info("Starting clustering preparations...")
    cur.execute("DROP TABLE IF EXISTS distances")
    log.debug("Done dropping!")
    cur.execute("CREATE TABLE distances LIKE alignments")
    cur.execute("INSERT INTO distances SELECT * FROM alignments")
    log.debug("Done filling distances!")

    # We prepare the 3 main dictionaries for the clustering. We create
    # one Node for every seq_id in the alignment table. We calculate
    # the nearest neighbour of the node to create the distance and
    # neighbour dicts. We also have to manually keep track of the
    # node_ids to save all nodes properly.

    node_id = 0
    dctClusters = {}
    dctNeig = {}
    dctDist = {}

    cur.execute("SELECT DISTINCT seq_id1 AS seq FROM distances")
    data = list(cur.fetchall())
    cur.execute("SELECT DISTINCT seq_id2 AS seq FROM distances")
    data.extend(list(cur.fetchall()))
    data = list(set(data))

    for i in data:
        seq = i[0]
        node_id += 1
        dctClusters[seq] = Node(seq, None, None, 0, node_id)
        cur.execute("SELECT seq_id1, seq_id2, evalue FROM distances "
                    "WHERE evalue=( SELECT MIN(evalue) FROM distances "
                    "WHERE seq_id1=%s or seq_id2=%s) "
                    "AND seq_id1=%s or seq_id2=%s LIMIT 1",
                    (seq, seq, seq, seq))
        seq_id1, seq_id2, eva = cur.fetchone()
        if seq_id1 == seq:
            dctNeig[seq] = seq_id2
        else:
            dctNeig[seq] = seq_id1
        dctDist[seq] = eva

    log.info("Done with preparations...")
    log.debug("Starting with %s individual nodes...",
              len(dctClusters.keys()))

    # The agglomeration runs as long as there are free nodes or if we
    # reach a defined threshold. Agglomerating without threshold
    # results in a single tree! In our case we have to set a threshold.
    # On every iteration we merge two nodes, delete the data of the
    # second node update the data of the fist node depending on the
    # clustering algorithm. We also need to update the neighbour and
    # distance dicts.

    log.info("Starting agglomeration...")

    dctClusterSize = {}
    fltEvalue = 1e-181

    for iteration in xrange(len(dctDist.keys())-1):
        if iteration % 10000 == 0:
            log.debug("Starting iteration %s", iteration)

        min_dist = min(dctDist.values())
        if min_dist >= threshold:
            log.debug("Threshold %s reached with %s", threshold,
                      min(dctDist.values()))
            break
        elif min_dist > fltEvalue:
            dctClusterSize[fltEvalue] = max(
                [len(c)+1 for c in dctClusters.values()]
                )
            fltEvalue *= 10

        c1_ = min(dctDist, key=dctDist.get)
        c2_ = dctNeig[c1_]
        c1 = min(c1_, c2_)
        c2 = max(c1_, c2_)

        c1Cluster = dctClusters[c1]
        c2Cluster = dctClusters[c2]

        node_id += 1
        c1Cluster.parent_id = node_id
        c2Cluster.parent_id = node_id
        newCluster = Node(None, c1Cluster, c2Cluster, dctDist[c1_], node_id)
        dctClusters[c1] = newCluster
        del dctClusters[c2]

        # We have to update all evalues of c1 depending on the
        # clustering method we use. First we need to get all the
        # evalues from the database.
        # We also need to take care of missing seq_id combinations. If
        # the combination was not found in the database, we set the
        # evalue to 1. We update the existing row or insert a new row
        # if it wasn't present before.

        lstUpdate = []
        lstInsert = []

        cur.execute("SELECT LEAST(seq_id1,seq_id2), "
                    "GREATEST(seq_id1,seq_id2), evalue "
                    "FROM distances WHERE seq_id1 IN %s "
                    "OR seq_id2 IN %s", ((c1, c2), (c1, c2)))
        evas = {(r[0], r[1]): r[2] for r in cur.fetchall()}

        for key in dctClusters.keys():
            insert = False
            if (min(c1, key), max(c1, key)) not in evas:
                eva1 = 1
                insert = True
            else:
                eva1 = evas[(min(c1, key), max(c1, key))]

            if (min(c2, key), max(c2, key)) not in evas:
                eva2 = 1
            else:
                eva2 = evas[(min(c2, key), max(c2, key))]

            new_eva = max(eva1, eva2)
            if new_eva != eva1:
                seq_id1 = min(c1, key)
                seq_id2 = max(c1, key)
                if insert:
                    lstInsert.append((seq_id1, seq_id2, new_eva))
                else:
                    lstUpdate.append((new_eva, seq_id1, seq_id2))

        cur.executemany("INSERT INTO distances (seq_id1,seq_id2,evalue) "
                        "VALUES (%s,%s,%s)", lstInsert, False)
        cur.executemany("UPDATE distances SET evalue=%s "
                        "WHERE seq_id1=%s AND seq_id2=%s", lstUpdate, False)

        # We have to delete all c2 entries from neighbours, distances
        # and the database.

        del dctDist[c2]
        del dctNeig[c2]
        cur.execute("DELETE FROM distances WHERE seq_id1=%s OR seq_id2=%s",
                    (c2, c2), False)

        # We get the new nearest neighbour for c1 from the database
        # and update the dictionaries. We also update all entries
        # which pointed to c2 before

        cur.execute("SELECT seq_id1, seq_id2, evalue FROM distances "
                    "WHERE seq_id1=%s OR seq_id2=%s ORDER BY evalue ASC "
                    "LIMIT 1", (c1, c1), False)
        cur.commit()
        res = cur.fetchone()

        if res is not None:
            seq_id1, seq_id2, eva = res

            if seq_id1 == c1:
                dctNeig[c1] = seq_id2
            else:
                dctNeig[c1] = seq_id1
            dctDist[c1] = eva
        else:
            del dctDist[c1]
            del dctNeig[c1]

        for key in dctNeig.keys():
            if dctNeig[key] == c2 and key != c1:
                dctNeig[key] = c1

            if dctNeig[key] == c1:
                seq_id1 = min(c1, key)
                seq_id2 = max(c1, key)

                cur.execute("SELECT seq_id1, seq_id2, evalue FROM "
                            "distances WHERE seq_id1=%s OR seq_id2=%s "
                            "ORDER BY evalue ASC "
                            "LIMIT 1", (key, key))
                res = cur.fetchone()
                if res:
                    seq_id1, seq_id2, eva = res
                    if seq_id1 == key:
                        dctNeig[key] = seq_id2
                    else:
                        dctNeig[key] = seq_id1
                    dctDist[key] = eva
                else:
                    if key in dctNeig:
                        del dctNeig[key]
                    if key in dctDist:
                        del dctDist[key]
    cur.close()
    log.info("Done agglomeration!")
    return dctClusters


def save(dctClusters):
    """
    Takes clusters created by :func:`.agglomerate` and saves them to
    the working database.

    Args:
        dctClusters (dict): A dictionary filled with :class:`.Node`
            objects representing the clusters, created by
            :func:`.agglomerate`.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    log.info("Saving cluster data...")
    cluster_id = 1
    lstClusters = []
    cur.execute("TRUNCATE TABLE nodes")

    cur.execute("SHOW INDEX FROM nodes WHERE Column_name=%s",
                ("cluster_id",))
    if len(cur.fetchall()) > 0:
        cur.execute("ALTER TABLE nodes DROP INDEX cluster_id")

    cur.execute("TRUNCATE TABLE clusters")
    cur.execute("TRUNCATE nodes_seqs_mapper")

    cur.execute("SHOW INDEX FROM nodes_seqs_mapper WHERE Column_name=%s",
                ("node_id",))
    if len(cur.fetchall()) > 0:
        cur.execute("ALTER TABLE nodes_seqs_mapper DROP INDEX node_id")

    # Wee need to increase Pythons default recursion limit to process
    # large clusters!
    sys.setrecursionlimit(10000)

    for key in dctClusters.keys():
        lstClusters.append((cluster_id, dctClusters[key].nodes))
        log.debug("Cluster {}: {} nodes".format(
            cluster_id, dctClusters[key].nodes))
        lstNodes = [n.to_tuple(cluster_id)
                    for n in dctClusters[key].get_nodes()]
        cur.executemany("INSERT INTO nodes (node_id,cluster_id,parent_id, "
                        "left_id,right_id,seq_id,dist,depth,size) "
                        "VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s)", lstNodes)
        lstMap = [
            (node.node_id, leaf.seq_id)
            for node in dctClusters[key].get_nodes()
            for leaf in node.get_leafs()
            if leaf.seq_id is not None
            ]
        lstChunks = [lstMap[i:i+1000] for i in xrange(0, len(lstMap), 1000)]
        for chunk in lstChunks:
            cur.executemany("INSERT INTO nodes_seqs_mapper (node_id, seq_id) "
                            "VALUES (%s,%s)", chunk)
        del lstMap
        del lstNodes
        cluster_id += 1

    cur.executemany("INSERT INTO clusters (cluster_id,nodes) "
                    "VALUES (%s,%s)", lstClusters)
    cur.execute("ALTER TABLE nodes ADD INDEX (cluster_id)")
    cur.execute("ALTER TABLE nodes_seqs_mapper ADD INDEX (node_id)")
    log.info("Done saving cluster data!")

if __name__ == "__main__":
    pass
