#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``createdb.py``
-----------------
This module creates the final databases in MySQL and SQLite format.
Additionally the patterns get an annotation with EC numbers, accessions,
and proposed function based on a consensus EC number.
"""
import re
import os
import sys
import json
import logging
import logging.config

import brepslib as bl

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

lstFields = ["breps_id", "ecs", "prop_ec", "prop_func", "accs", "num_ecs",
             "evalue", "node_id", "cluster_id"]

lstPatFields = ["type", "breps_id", "pat_re", "pat_ps", "pat_sub",
                "comp", "min_len", "max_len"]

lstRelFields = ["year", "serial", "version", "date", "uniprot", "enzymes",
                "nonenzymes", "patterns",  "extended", "subpatterns",
                "protocol"]

strExe = "INSERT INTO {} ({}) VALUES ({})"


def pattern_len(strPS):
    """
    Calculates the minimum and maximum length of a pattern.

    Args:
        strPS (string): Pattern in PROSITE format.

    Returns:
        tuple: Tuple with two values for minimu and maximum pattern
        length.
    """
    lstPat = strPS.split("-")
    intMin = 0
    intMax = 0
    for p in lstPat:
        if "," in p:
            m = re.search("\((\d*)\,(\d*)\)", p)
            intMin += int(m.group(1))
            intMax += int(m.group(2))
        elif "(" in p:
            m = re.search("\((\d*)\)", p)
            intMin += int(m.group(1))
            intMax += int(m.group(1))
        else:
            intMin += 1
            intMax += 1
    return intMin, intMax


def count_positions(strPS):
    """
    Calculates the number of positions in a pattern.

    Args:
        strPS (string): Pattern in PROSITE format.

    Returns:
        int: Number of positions.
    """
    lstPat = strPS.split("-")
    lstCount = [[0, 0], [0, 0], [0, 0]]
    for p in lstPat:
        mn, mx = pattern_len(p)
        if "x" in p:
            lstCount[0][0] += mn
            lstCount[0][1] += mx
        elif "[" in p:
            lstCount[1][0] += mn
            lstCount[1][1] += mx
        else:
            lstCount[2][0] += mn
            lstCount[2][1] += mx
    return lstCount


def subpatterns(strPS, intGaplen=100):
    """
    Generates sub-patterns for a given pattern.

    The sub-patterns are calculates based on a given gap length for
    variable positions in the pattern. A sub-pattern needs at least 8
    positions (just like a regular pattern).

    .. note:: Sub-patterns are currently not used by BrEPS 2.0.

    Args:
        strPS (str): Pattern in PROSITE format.
        intGaplen (int): Minimum length of a gap to cut the pattern
            (Default 100).

    Returns:
        list: A list of sub-patterns in PROSITE format for the given
        pattern.
    """
    lstSubPatterns = []
    strPat = strPS
    lstMatches = re.findall("x\(([0-9]{1,4}),([0-9]{1,4})\)", strPS)

    # Find all gaps in the pattern and check if the gap length exceedes
    # the defined gap length. We only save subpatterns with >= 8
    # semi-conserved or conserved positions.

    fltComp = bl.complexity(strPS)
    fh = open("crap_subpattern.txt", "a")
    for mn, mx in lstMatches:
        mn = int(mn)
        mx = int(mx)

        if mx-mn > intGaplen:
            lstSubs = strPat.split("x({mn},{mx})".format(mn=mn, mx=mx))
            strSub = lstSubs[0].strip("-")
            cnt = count_positions(strSub)
            cnt = cnt[1][0] + cnt[2][0]
            if cnt >= 8 and bl.complexity(strSub) < fltComp:
                lstSubPatterns.append(strSub)
            else:
                # Add subpattern to file
                fh.write(strSub)
                fh.write("\n")
            strPat = lstSubs[1].strip("-")

    # If we got patterns, add the last part to the list (if long enough)
    if lstSubPatterns:
        cnt = count_positions(strPat.strip("-"))
        cnt = cnt[1][0] + cnt[2][0]
        if cnt >= 8 and bl.complexity(strPat.strip("-")) < fltComp:
            lstSubPatterns.append(strPat.strip("-"))
        else:
            # Add subpattern to file
            fh.write(strSub)
            fh.write("\n")
    fh.close()
    return lstSubPatterns


def match_ecs(strEc1, strEc2):
    """
    Creates a partial match of two EC numbers.

    Example: 1.1.1.1 and 1.1.1.2 create 1.1.1

    Args:
        strEc1 (str): The first EC number.
        strEc2 (str): The second EC number.

    Returns:
        str: the overlapping part of both EC numbers or ``None`` if no
        match.
    """
    lstEc1 = [x for x in strEc1.split(".") if x != "-"]
    lstEc2 = [x for x in strEc2.split(".") if x != "-"]
    lstMatch = []
    for i in xrange(4):
        if i < len(lstEc1) and i < len(lstEc2) and lstEc1[i] == lstEc2[i]:
            lstMatch.append(lstEc1[i])
        else:
            break

    if lstMatch:
        return ".".join(lstMatch)
    else:
        return None


def get_consensus(strEcs):
    """
    Creates a consensus EC number from multiple EC numbers.

    The lowest EC number class level where all ECs match. Result can
    be a partial EC number (like 1.1.1.-).

    Args:
        strEcs (str): Multiple EC numbers separated with , (comma),
            multiple domains separated by & (ampersand).

    Returns:
        list: A list with all consensus EC numbers.
    """
    lstGroups = [g.split("&") for g in strEcs.split(", ")]
    setEcs = {e for g in lstGroups for e in g}
    dctGroups = {}
    lstMatches = []
    lstEcs = []
    lstAnd = []
    # Create all overlapping EC variations
    for ec1 in setEcs:
        for ec2 in setEcs:
            strMatch = match_ecs(ec1, ec2)
            if strMatch is not None and strMatch not in lstMatches:
                lstMatch = strMatch.split(".")
                lstMatch.extend(["-"]*(4-len(lstMatch)))
                strMatchExt = ".".join(lstMatch)
                lstMatches.append(strMatch)
                dctGroups[strMatchExt] = []
                for lstEcs in lstGroups:
                    for strEc in lstEcs:
                        if (
                            (strMatchExt.count("-") > 0 and
                             strEc.startswith(strMatch + ".")) or
                            (strMatchExt.count("-") == 0 and
                             strEc == strMatch)):
                            dctGroups[strMatchExt].append(lstEcs)
                            break
    lstEcs = []
    lstAnd = []
    # Sort by EC quality and group coverage
    lstSorted = [a for a in
                 sorted(dctGroups, key=lambda k: 4-k.count("-"),
                        reverse=True)]
    lstSorted = [a for a in
                 sorted(lstSorted, key=lambda k: len(dctGroups[k]),
                        reverse=True)]
    # Get best ECs
    intCov = max([len(dctGroups[x]) for x in dctGroups])
    intQuali = 0
    for g in lstSorted:
        hits = [x for x in lstAnd if dctGroups[x] == dctGroups[g]]
        if (4-g.count("-") > intQuali and intCov == len(dctGroups[g])):
            intQuali = 4-g.count("-")
            lstAnd = [g]
        elif len(hits) == len(lstAnd) and (g in setEcs):
            lstAnd.append(g)

    # Check if we got everything covered
    lstEcs.append(lstAnd)
    lstCov = [i for strMax in lstAnd
              for i in dctGroups[strMax] if i in lstGroups]
    lstUncov = [i for i in lstGroups if i not in lstCov]

    if lstUncov:
        # Get missing consensus if needed
        strEcs = ", ".join(["&".join(l) for l in lstUncov])
        lstEcs.extend(get_consensus(strEcs))
    return lstEcs


def prepare_db():
    """
    Prepares the MySQL and SQLite database.

    Prepares SQLite and MySQL databases for final release.

    """
    # Create SQLite database
    log.info("Creating SQLite database")
    cur = bl.Dbase("sqlite", "{}.db".format(dctCfg["final_db"]))
    cur.execute("DROP TABLE IF EXISTS annotation")
    cur.execute("CREATE TABLE annotation (breps_id INTEGER PRIMARY KEY,"
                "ecs,prop_ec,prop_func,accs,num_ecs,evalue,node_id,"
                "cluster_id)")
    cur.execute("DROP TABLE IF EXISTS patterns")
    cur.execute("CREATE TABLE patterns (pat_id INTEGER PRIMARY KEY,"
                "type,breps_id,pat_re,pat_ps,pat_sub,comp,min_len,"
                "max_len)")
    cur.execute("DROP TABLE IF EXISTS releaseinfo")
    cur.execute("CREATE TABLE releaseinfo (res_id INTEGER PRIMARY KEY,"
                "year,serial,version,date,uniprot,enzymes, "
                "nonenzymes, patterns, extended,subpatterns,protocol)")
    cur.execute("DROP TABLE IF EXISTS results")
    cur.close()

    # Create MySQL database
    log.info("Truncating MySQL database")

    # TODO: Add db creation from file!
    cur = bl.Dbase("mysql", dctCfg["final_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.execute("TRUNCATE TABLE patterns")
    cur.execute("TRUNCATE TABLE annotation")
    cur.execute("TRUNCATE TABLE results")
    cur.close()


def save(strTable, lstData, lstFields):
    """
    Saves data to MySQL and SQLite.

    Args:
        strTable (str): Table name to save in.
        lstData (list): List of tuples with data to save.
        lstFields (list): List of field names for query generation.
    """
    strQuery = strExe.format(strTable, ",".join(lstFields),
                             ",".join(["%s"] * len(lstFields)))
    log.info("Saving {} as SQLite".format(strTable))
    cur = bl.Dbase("sqlite", "{}.db".format(dctCfg["final_db"]))
    cur.executemany(strQuery, lstData)
    cur.close()
    log.info("Saving {} as MySQL".format(strTable))
    cur = bl.Dbase("mysql", dctCfg["final_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.executemany(strQuery, lstData)
    cur.close()
    # TODO: export MySQl file


def prepare_data():
    """
    Prepares pattern data for final database creation.

    Returns:
        tuple: Tuple containing two lists of tuples with annotation and
        pattern data for saving.

    """
    # We want to connect to our source database to get pattern data.
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    log.info("Preparing pattern data...")

    # We want to join the data of different tables into one result. Also
    # we only fetch patterns with positive prediction value of 1!
    cur.execute("SELECT p.pat_id, p.pattern_re AS pat_re, p.pattern_ps AS "
                "pat_ps,p.complexity AS comp, p.node_id, n.cluster_id, "
                "GROUP_CONCAT(DISTINCT s.ecs SEPARATOR ', ') AS ecs, "
                "COUNT(DISTINCT s.ecs) AS num_ecs, p.min_len, p.max_len, "
                "n.dist AS evalue FROM patterns p "
                "LEFT JOIN nodes n ON n.node_id=p.node_id "
                "LEFT JOIN nodes_seqs_mapper m ON m.node_id=p.node_id "
                "LEFT JOIN sequences s ON m.seq_id=s.seq_id "
                "WHERE (p.tp/(p.fp+p.tp)) = 1.0"
                "GROUP BY p.pat_id")

    lstPats = cur.fetchall()

    dctFunction = {}
    lstAnnotation = []
    lstPatterns = []

    lstFields = ["pat_id", "pat_re", "pat_ps", "comp", "node_id", "cluster_id",
                 "ecs", "num_ecs", "min_len", "max_len", "evalue"]

    for tplPat in lstPats:
        row = {lstFields[_]: tplPat[_] for _ in xrange(len(tplPat))}

        # We create a consensus EC number from all ECs associated with
        # the pattern and use the IUBMB nomenclature database to fetch
        # the function of the consensus EC number(s).
        row["ecs"] = row["ecs"].replace(";", "&")
        lstConsensus = get_consensus(row["ecs"])
        row["prop_ec"] = " OR ".join([" AND ".join(a) for a in lstConsensus])
        lstEcs = [ec for lst in lstConsensus for ec in lst]
        for ec in lstEcs:
            if ec not in dctFunction:
                if "-" in ec:
                    lstClass = [
                        int(x) if x != "-" else 0 for x in ec.split(".")
                        ]
                    cur.execute("SELECT heading FROM class WHERE "
                                "class=%s AND subclass IN (%s,%s) AND "
                                "subsubclass IN (%s,%s) ORDER BY "
                                "class, subclass, subsubclass",
                                (lstClass[0], 0, lstClass[1], 0, lstClass[2]))
                    entry = cur.fetchall()
                    if entry:
                        dctFunction[ec] = " > ".join([e[0] for e in entry])
                    else:
                        dctFunction[ec] = "Undefined ({})".format(ec)
                else:
                    cur.execute("SELECT accepted_name FROM entry WHERE "
                                "ec_num=%s", (ec,))
                    entry = cur.fetchone()
                    if entry:
                        dctFunction[ec] = entry[0][0].capitalize()
                        dctFunction[ec] += entry[0][1:]
                    else:
                        dctFunction[ec] = ec
        lstOr = []
        for lstCon in lstConsensus:
            lstAnd = []
            for ec in lstCon:
                lstAnd.append(dctFunction[ec])
            lstOr.append(" AND ".join(lstAnd))
        row["prop_func"] = " OR ".join(lstOr)

        row["pat_sub"] = bl.ps2regex(bl.subpattern(row["pat_ps"]))

        # We fetch all associated accessions of the pattern and
        # concaternate the using a ','. A simple GROUP_CONCAT() wouldn't
        # fetch all accessions.
        cur.execute("SELECT s.accession, s.ecs FROM nodes_seqs_mapper m "
                    "LEFT JOIN sequences s ON s.seq_id=m.seq_id "
                    "WHERE m.node_id=%s", (row["node_id"],))
        lstAccs = cur.fetchall()
        row["accs"] = ", ".join([_[0] for _ in lstAccs])
        dctEcs = {}
        for tplAcc in lstAccs:
            if tplAcc[1] not in dctEcs:
                dctEcs[tplAcc[1]] = 0
            dctEcs[tplAcc[1]] += 1

        # We want to count the number of sequences/accessions for every
        # EC number.
        row["ecs"] = ", ".join(
            ["{} ({})".format(x, dctEcs[x]) for x in dctEcs]
            )
        row["ecs"] = row["ecs"].replace(";", "& ")

        # Finally insert the compressed data into the source databases
        # and save the new row id as we drop some pattern with ppv < 1.
        tplData = (row["pat_id"], row["ecs"], row["prop_ec"], row["prop_func"],
                   row["accs"], row["num_ecs"], row["evalue"], row["node_id"],
                   row["cluster_id"])
        lstAnnotation.append(tplData)
        tplData = ("std", row["pat_id"], row["pat_re"], row["pat_ps"],
                   row["pat_sub"], row["comp"], row["min_len"], row["max_len"])
        lstPatterns.append(tplData)

        # We also want to get the extended pattern, if the extended
        # patterns PPV is 1 as well and the complexity is lower.
        cur.execute("SELECT pattern_re, pattern_ps, complexity, min_len, "
                    "max_len FROM patterns_extended WHERE pat_id=%s AND "
                    "(tp/(tp+fp))=1.0 AND complexity > %s",
                    (row["pat_id"], row["comp"]))
        lstExt = cur.fetchone()
        if lstExt:
            strSub = bl.ps2regex(
                bl.subpattern(lstExt[1]))
            tplData = ("ext", row["pat_id"], lstExt[0], lstExt[1],
                       strSub, 1-lstExt[2], lstExt[3], lstExt[4])
            lstPatterns.append(tplData)

    return (lstAnnotation, lstPatterns)

if __name__ == "__main__":
    prepare_data()
