#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``data.py``
-----------
This module handles the download and preparation of all needed data.

The ``data`` folder is associated to this module.
"""

import re
import os
import sys
import gzip
import json
import shutil
import ftplib
import logging
import datetime
import logging.config

import brepslib as bl

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

UNIPROT_FTP = dctCfg["uniprot_ftp"]
UNIPROT_DIR = dctCfg["uniprot_dir"]
SPROT_FLATFILE = "uniprot_sprot.dat.gz"
TREMBL_FLATFILE = "uniprot_trembl.dat.gz"

UNIREF_FTP = dctCfg["uniref_ftp"]
UNIREF_DIR = dctCfg["uniref_dir"]
UNIREF_FILE = "uniref50.xml.gz"

IUBMB_FTP = dctCfg["iubmb_ftp"]
IUBMB_DIR = dctCfg["iubmb_dir"]
IUBMB_FILE = "enzyme-data.sql.gz"

reExclude = re.compile(
    "(putative|hypothetical|fragment"
    "|probable|possible|potential)[^\w]", flags=re.I)
reLength = re.compile(r" ([1-9]\d{2}) AA| ([1-6]\d{3}) AA")
reAcc = re.compile(
    r"([OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]"
    "([A-Z][A-Z0-9]{2}[0-9]){1,2})")
reDate = re.compile(r"(\d{2}-\w{3}-\d{4}), sequence version \d*")
reEcs = re.compile(r"[1-6]\.[0-9\-]{1,2}\.[0-9\-]{1,2}\.[0-9\-]{1,3}")
reSeq = re.compile(r"([ARNDCQEGHILKMFPSTWYVOUBZX ]*)")
re_swissprot = re.compile(
    r"UniProtKB/Swiss-Prot Release ([0-9_]*) "
    "of (\d{2}-\w{3}-\d{4})"
    )


def get_release(strFile):
    """Extracts the release date from strFile.

    The file has to be in Uniprots reldate.txt format containing three
    lines. First line is the release identifier (<year>_<month>, eg.
    2015_12) of UniprotKB. The next two lines are the release idetifier
    for Swiss-Prot and TrEMBL data with corresponding release date
    (<day>-<month>-<year>, eg. 09-Dec-2015).

    Args:
        strFile (str): Path to the file containing the reldate data.

    Returns:
        tuple: A tuple containing date of the update in seconds and
        version identifier.
    """
    with open(strFile, "r") as fh:
        lastVersion = None
        lastUpdate = 0
        for line in fh.readlines():
            m = re.match(re_swissprot, line)
            if m is not None:
                lastVersion = m.group(1)
                lastDate = datetime.datetime.strptime(m.group(2), "%d-%b-%Y")
                lastUpdate = lastDate.strftime("%s")
    fh.close()
    return (lastUpdate, lastVersion)


def new_version():
    """
    Downloads current version of UniProt from the FTP server.

    Connects to the uniprot server via FTP to retrieve the
    ``reldate.txt`` file containing the release informations.

    Returns:
        boolean: True if there is a newer release available,
        False if not.
    """
    log.info("Checking UniProt version...")
    if os.path.isfile("{}/data/reldate.txt".format(strPath)):
        fh = open("{}/data/reldate.txt".format(strPath), "rb+")
        lines = fh.readlines()

        if len(lines) == 0:
            lastUpdate = 0
        else:
            lastUpdate, lastVersion = get_release(
                "{}/data/reldate.txt".format(strPath)
                )
    else:
        fh = open("{}/data/reldate.txt".format(strPath), "wb")
        lastUpdate = 0

    # We connect to the uniprot ftp server and move to the directory
    # containing the database. We retrieve the reldate.txt to check for
    # a new release of uniprot, save the file and return True if there
    # is a new release.

    ftp = ftplib.FTP(UNIPROT_FTP)
    response = ftp.login()
    if not response.split()[0] == "230":
        log.error("Could not connect to ftp server %s", UNIPROT_FTP)
        raise IOError("Could not connect to ftp server {}".format(UNIPROT_FTP))
        sys.exit()

    response = ftp.cwd(UNIPROT_DIR)
    if not response.split()[0] == "250":
        log.error("Could not move to directory %s", UNIPROT_DIR)
        raise IOError("Could not move to directory {}".format(UNIPROT_FTP))
        sys.exit()

    fh.seek(0)
    ftp.retrbinary("RETR %s" % "reldate.txt", fh.write)
    fh.truncate()
    fh.close()
    ftp.quit()

    newUpdate, newVersion = get_release("{}/data/reldate.txt".format(strPath))

    if newUpdate > lastUpdate:
        log.info("Found new UniProt version %s!", newVersion)
        return True
    else:
        log.info("No new UniProt version found!")
        return False


def download(strServer, strFolder, strFile):
    """
    Retrieves a file either from ftp or a local path.

    Args:
        strServer (str): FTP Server to connect to, ``None`` for local
            file.
        strFolder (str): Path to the file, either FTP or local.
        strFile (str): File to download from FTP sever or to copy from.

    Returns:
        str: Path to the downloaded file.

    Raises:
        IOError: If could not connect to ftp or folder does not exist.
    """
    # If no FTP server is given, the file will be copied from the given
    # folder. Otherwise the file will be retrieved from FTP. Absolute
    # path to the folder needs to be given. Existing files will be
    # overwritten!
    if strServer is None:
        log.info("Copying {} from {}...".format(strFile, strFolder))
        if not os.path.isfile("{}{}".format(strFolder, strFile)):
            log.error("File not found: {}{}".format(strFolder, strFile))
            raise IOError(
                "File not found: {}{}".format(strFolder, strFile)
                )
            sys.exit()
        shutil.copyfile(
            "{}{}".format(strFolder, strFile),
            "{}/data/{}".format(strPath, strFile))
    else:
        log.info("Downloading {} from {}...".format(strFile, strServer))
        ftp = ftplib.FTP(strServer)

        response = ftp.login()
        if not response.split()[0] == "230":
            log.error("Could not connect to ftp server {}".format(strServer))
            raise IOError(
                "Could not connect to ftp server {}".format(strServer)
                )
            sys.exit()

        response = ftp.cwd(strFolder)
        if not response.split()[0] == "250":
            log.error("Could not move to directory {}".format())
            raise IOError("Could not move to directory {}".format(strFolder))
            sys.exit()

        fh = open("{}/data/{}".format(strPath, strFile), "wb")

        log.debug("Starting download of {}...".format(strFile))
        ftp.retrbinary("RETR %s" % strFile, fh.write)

        fh.close()
        ftp.quit()
    return "{}/data/{}".format(strPath, strFile)


def process_date(strDate):
    """
    Parser helper function. Converts ``dd-mmm-yyyy`` to timestamp from
    a string.

    Args:
        strDate (str): Date in ``dd-mm-yyyy`` format.

    Returns:
        datetime: datetime object for the corresponding date.
    """
    m = reDate.search(strDate)
    return datetime.datetime.strptime(
        m.group(1), "%d-%b-%Y")


def process_acc(strAcc):
    """
    Parser helper function. Returns the primary accession from a string.

    More infos about primary accessions here:
    (see http://web.expasy.org/docs/userman.html#AC_line)

    Args:
        strAcc (str): One or more comma separated accessions.

    Returns:
        str: First (primary) accession.
    """
    m = reAcc.search(strAcc)
    if m is not None:
        return m.group(1)


def process_seq(strSeq):
    """
    Parser helper function. Replaces whitespaces from sequence string.
    """
    m = reSeq.search(strSeq)
    if m is not None:
        return m.group(1).replace(" ", "")


def process_ecs(strEcs):
    """
    Parser helper function. Returns all found ec numbers as a string
    separated by a semicolon
    """
    lstEcs = reEcs.findall(strEcs)
    lstEcs = list(set(lstEcs))
    return ";".join(lstEcs)


class basic_parser(object):
    """
    Basic parser for UniProt files.

    The basic parser saves accession and sequence of every UniProt entry
    with a specific length defined in ``reLength``.
    This is just a proof on concept for more advanced parsers. Methods
    ``_require()``, ``_exclude()`` and ``_process()`` are supposed to be
    overwritten in advanced parsers.

    Attributes:
        lstFields (list): Two character line identifiers to save while
            parsing the file.

    """
    lstFields = ["AC", "  ", "SQ"]

    def __init__(self, strFile):
        """
        Inits the parser for a given file ``strFile``.

        Args:
            strFile (str): The file to parse.

        Raises:
            IOError: If file in ``strFile`` is not found.
        """

        super(basic_parser, self).__init__()
        self.strFile = strFile

    def __str__(self):
        pass

    def __len__(self):
        return len(self.lstData)

    # TODO: merge _require and _exclude functions
    def _require(self, dctEntry):
        """
        Checks if all requirements are met for a data entry.

        Args:
            dctEntry (dict): A dictionary containing all gathered data
                fileds.

        Returns:
            boolean: ``True`` if the conditions are met,
            ``False`` otherwise.
        """
        return bool(reLength.search(dctEntry["SQ"]))

    def _exclude(self, dctEntry):
        """
        Checks if a data entry needs to be excluded.

        Args:
            dctEntry (dict): A dictionary containing all gathered data
                fileds.

        Returns:
            boolean: ``True`` if the entry needs to be excluded,
            ``False`` otherwise.
        """
        return False

    def _process(self, dctEntry):
        """
        Processes the data entry using helper functions.

        Args:
            dctEntry(dict): A dictionary containing all gathered data
                fileds.

        Returns:
            dict: A dictionary with processed data fields.
        """
        dctEntry["accession"] = process_acc(dctEntry["AC"])
        dctEntry["sequence"] = process_seq(dctEntry["  "])
        dctEntry["seq_len"] = len(dctEntry["  "])
        dctEntry["source"] = "swissprot"
        del dctEntry["AC"]
        del dctEntry["  "]
        del dctEntry["SQ"]
        return dctEntry

    def get_fields(self):
        return self.lstData[0].keys()

    def as_list(self):
        return [i.values() for i in self.lstData]

    def as_dict(self):
        return self.lstData

    def parse(self, openwith=gzip.GzipFile):
        """
        Starts the parsing of the given UniProt file.

        The parser works in three steps for every entry in the UniProt
        Flatfile.

            #. Check if requirements are met.
            #. Check if the data needs to be excluded.
            #. Process the data.

        Successfully prepared datasets are directly saved to database.

        Args:
            openwith (function): Function to open the file. Needs to
                work just like Pythons ``open()`` function.
                (Default: ``gzip.GzipFile``)
        """
        self.lstData = []
        try:
            fh = openwith(self.strFile, "rb")
        except IOError, e:
            log.error("File %s not found!", self.strFile)
            raise e
        log.debug("Start parsing of %s using %s...",
                  self.strFile.split("/")[-1], type(self).__name__)
        dctEntry = {}
        for line in fh:
            line = line.rstrip()
            if line == "//":
                # Step 1: check requirements
                if not self._require(dctEntry):
                    dctEntry = {}
                    continue

                # Step 2: exclude data
                if self._exclude(dctEntry):
                    dctEntry = {}
                    continue

                # Step 3: process data
                self.lstData.append(self._process(dctEntry))
                dctEntry = {}

            elif line[:2] in self.lstFields:
                if line[:2] in dctEntry:
                    dctEntry[line[:2]] += line[2:].lstrip()
                else:
                    dctEntry[line[:2]] = line[2:].lstrip()
                if len(self.lstData) >= 500000:
                    self.save()
        self.save()
        fh.close()

    def save(self):
        """
        Saves the parsed data saved in ``self.lstData`` to database.

        Data is saved to the ``sequences`` table of the working
        database. The filed names are automatically fetched from the
        dictionary keys.
        """
        cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])

        lstResults = []
        strFields = ",".join(self.lstData[0].keys())

        strPlaceholders = ",".join(["%s"]*len(self.lstData[0].keys()))
        strQuery = ("INSERT INTO sequences ({}) "
                    "VALUES ({})".format(strFields,
                                         strPlaceholders))

        for data in self.lstData:
            lstResults.append(data.values())
        log.debug("Saving %s datasets...", len(lstResults))
        self.lstData = []

        lstChunks = [lstResults[i:i+500]
                     for i in xrange(0, len(lstResults), 500)]
        for chunk in lstChunks:
            cur.executemany(strQuery, chunk)

        cur.close()


class original_parser(basic_parser):
    """
    Parser as used by the original BrEPS protocol.

    This parser inherits basic functionality from
    :class:`.basic_parser`.

    Requirements:

        - Sequence length 100-6999 aa (see ``reLength``).

    Exclusion:

        - Keywords in DE field: putative, hypothetical, fragment,
          probable, possible, potential.

    Entries are saved as enzymes (filed enzyme=1) if DE field contains
    a valid EC number.
    """
    lstFields = ["AC", "  ", "DE", "SQ"]

    def __init__(self, strFile):
        cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])
        cur.execute("TRUNCATE TABLE sequences")
        cur.close()
        super(original_parser, self).__init__(strFile)

    def _process(self, dctEntry):
        dctResult = {}
        dctResult["accession"] = process_acc(dctEntry["AC"])
        dctResult["ecs"] = process_ecs(dctEntry["DE"])
        dctResult["sequence"] = process_seq(dctEntry["  "])
        dctResult["seq_len"] = len(dctResult["sequence"])
        dctResult["enzyme"] = bool(reEcs.search(dctEntry["DE"]))
        dctResult["source"] = "swissprot"
        del dctEntry
        return dctResult

    def _require(self, dctEntry):
        return bool(reLength.search(dctEntry["SQ"]))

    def _exclude(self, dctEntry):
        return bool(reExclude.search(dctEntry["DE"]))


class sprot_parser(original_parser):
    """
    Parser for the advanced data preparation in the BrEPS 2.0 protocol.

    This parser inherits basic functionality from
    :class:`.original_parser` and is intended to be used on Swiss-Prot
    flatfiles.

    Requirements:

        - Sequence length 100-6999 aa (see ``reLength``).
        - Protein existence on protein level (PE 1).

    Exclusion:

        - Keywords in DE field: putative, hypothetical, fragment,
          probable, possible, potential.

    Entries are saved as enzymes (filed enzyme=1) if DE field contains
    a valid EC number and PE = 1. Non enzymes are saved independent from
    protein existence.
    """
    lstFields = ["AC", "  ", "DE", "SQ", "PE"]

    def __init__(self, strFile):
        cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])
        cur.execute("TRUNCATE TABLE sequences")
        cur.close()
        super(sprot_parser, self).__init__(strFile)

    def _require(self, dctEntry):
        # Exclude all enzymes without PE 1
        blnEnzyme = bool(reEcs.search(dctEntry["DE"]))
        if blnEnzyme and dctEntry["PE"][0] != "1":
            return False
        return super(sprot_parser, self)._require(dctEntry)

    def _process(self, dctEntry):
        dctResult = {}
        dctResult["accession"] = process_acc(dctEntry["AC"])
        dctResult["ecs"] = process_ecs(dctEntry["DE"])
        dctResult["sequence"] = process_seq(dctEntry["  "])
        dctResult["seq_len"] = len(dctResult["sequence"])
        dctResult["enzyme"] = bool(reEcs.search(dctEntry["DE"]))
        dctResult["source"] = "swissprot"
        del dctEntry
        return dctResult


class advanced_parser(basic_parser):
    """
    Parser for the advanced data preparation in the BrEPS 2.0 protocol.

    This parser inherits basic functionality from
    :class:`.basic_parser` and is intended to be used on TrEMBL
    flatfiles. The parser needs the dictionray ``dctRef`` containing
    TrEMBL accessions based on sequence similarity from a UniRef file.

    Requirements:

        - Sequence length 100-6999 aa (see ``reLength``).
        - Cross-reference (RX) entry.

    Exclusion:

        - Accession is not similar to a Swiss-Prot sequence based on
          UniRef data in ``dctRef``.
        - Sequence length does not matches original length +- 25%

    Entries are saved with the corresponding Swiss-Prot accession in the
    source field.
    """
    lstFields = ["AC", "  ", "SQ", "RX", "DT"]

    def __init__(self, strFile, dctRef):
        self.dctRef = dctRef
        cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                       dctCfg["db_user"], dctCfg["db_pass"])
        cur.execute("SELECT accession, ecs, seq_len "
                    "FROM sequences WHERE enzyme=1")
        self.dctEcs = {s[0]: [s[1], s[2]] for s in cur.fetchall()}
        cur.close()
        super(advanced_parser, self).__init__(strFile)

    def _process(self, dctEntry):
        dctResult = {}
        dctResult["accession"] = process_acc(dctEntry["AC"])
        strEcs = ";".join(
            [self.dctEcs[a][0] for a in self.dctRef[dctResult["accession"]]]
            )
        dctResult["ecs"] = strEcs
        dctResult["sequence"] = process_seq(dctEntry["  "])
        dctResult["seq_len"] = len(dctResult["sequence"])
        dctResult["source"] = ";".join(self.dctRef[dctResult["accession"]])
        return dctResult

    def _require(self, dctEntry):
        if "RX" not in dctEntry:
            return False
        return super(advanced_parser, self)._require(dctEntry)

    def _exclude(self, dctEntry):
        if process_acc(dctEntry["AC"]) not in self.dctRef:
            return True
        else:
            acc = self.dctRef[process_acc(dctEntry["AC"])][0]
            intRange = self.dctEcs[acc][1]*0.25
            intMaxLen = self.dctEcs[acc][1] + intRange
            intMinLen = self.dctEcs[acc][1] - intRange
            intLen = len(process_seq(dctEntry["  "]))
            if intLen < intMinLen or intLen > intMaxLen:
                return True


def parse_uniref(strFile, openwith=gzip.GzipFile):
    """
    Very simple parser for UniRef XML files.

    The parser function basically fetches accessions in the
    ``representativeMember`` tag and saves corresponding accessions in
    the ``member`` tag.

    .. todo::
        Replace the function with a proper XML parser!

    Args:
        strFile (str): Path to the UniRef XML file to parse.
        openwith (function): Function to open the file. Needs to
            work just like Pythons ``open()`` function.
            (Default: ``gzip.GzipFile``)

    Returns:
        dict: Dictionary containing Trembl accession as key and
        corresponding SwissProt accession.

    Raises:
        IOError: UniRef file doesn't exist.
    """

    reRef = re.compile("<representativeMember>.*UniProtKB accession.*\"([OPQ]"
                       "[0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}"
                       "[0-9]){1,2})\"", flags=re.I)
    reMember = re.compile("<member>.*UniProtKB accession.*\"([OPQ][0-9]"
                          "[A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}"
                          "[0-9]){1,2})\"", flags=re.I)

    try:
        fh = openwith(strFile)
    except IOError, e:
        log.error("File %s not found!", strFile)
        raise e
    log.debug("Start UniRef parser...")
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    # Get all enzyme accessions
    cur.execute("SELECT accession FROM sequences WHERE enzyme=1")
    lstAccs = [a[0] for a in cur.fetchall()]

    log.info("Parsing UniRef file...")
    dctRef = {}
    strEntry = ""
    for line in fh:
        strEntry += line.strip()
        if "</entry>" in line:
            subs = strEntry.split("</representativeMember>")
            m = reRef.search(subs[0])
            if bool(m) and m.group(1) in lstAccs:
                strRefAcc = m.group(1)
                members = subs[1].split("</member>")
                for sub in members:
                    n = reMember.search(sub)
                    if bool(n):
                        if n.group(1) not in dctRef:
                            dctRef[n.group(1)] = [strRefAcc]
                        else:
                            dctRef[n.group(1)].append(strRefAcc)
            strEntry = ""
    fh.close()
    cur.close()
    return dctRef


if __name__ == "__main__":
    # The file will be called from cmaster01 via ssh command. All file
    # downloads and copies need to be done via cmaster01.
    # For desktop mode this will be simply called.

    blnAdvanced = True

    strIubmbFile = download(IUBMB_FTP, IUBMB_DIR, IUBMB_FILE)
    strSwissprotFile = download(UNIPROT_FTP, UNIPROT_DIR,
                                SPROT_FLATFILE)

    # Load IUBMB data into working database
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.load_sql(strIubmbFile)
    cur.close()

    # TrEMBL and UniRef files are only needed for the advanced
    # protocol.
    if blnAdvanced:
        strTremblFile = download(UNIPROT_FTP, UNIPROT_DIR,
                                 TREMBL_FLATFILE)
        strUnirefFile = download(UNIREF_FTP, UNIREF_DIR,
                                 UNIREF_FILE)

        # Parse seed sequences from Swiss-Prot files
        objParser = sprot_parser(strSwissprotFile)
        objParser.parse()

        # Parse UniRef file
        dctRef = parse_uniref(strUnirefFile)

        # Parse additional sequences from TrEMBL
        objParser = advanced_parser(strTremblFile, dctRef)
        objParser.parse()
        # Parse additional sequences from Swiss-Prot
        objParser = advanced_parser(strSwissprotFile, dctRef)
        objParser.parse()

    # The original protocol simply parses the Swiss-Prot file for
    # enzymes and non-enzymes.
    else:
        objParser = original_parser(strSwissprotFile)
        objParser.parse()

    # As this runs separately from the core script, errors will not stop
    # the BrEPS pipeline. To avoid problems, we want to create a file
    # as proof of finish.
    os.system("touch {}/data/done.tmp".format(strPath))
