#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``pattern.py``
-----------------
This module clusters prepares FASTA files for the multiple sequence
alignments (MSA). The actual MSA is created using Clustal Omega. Based
on the MSA output the patterns are created and additional extended
patterns are created.

The ``pattern`` folder is associated to this module.
"""
import os
import re
import sys
import json
import glob
import time
import logging
import logging.config
import multiprocessing as mp

import brepslib as bl

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

CLUSTAL_DIR = "{}/clustal-current/bin".format(os.getcwd())
SEQ_DIR = "{}/pattern".format(os.getcwd())

lstAA = ["C", "S", "T", "P", "A", "G", "N", "D", "E", "Q", "H",
         "R", "K", "M", "I", "L", "V", "F", "Y", "W"]

# Gonnets PAM25 matrix derived from:
# Gonnet G, Cohen M, Benner S. Exhaustive matching of the entire
# protein sequence database. Science (80- ). 1992;256(5062):1443–1445.
# doi:10.1126/science.1604319.
lstPam250 = [
    [11.5, 0.1, -0.5, -3.1,  0.5, -2.0, -1.8, -3.2, -3.0, -2.4, -1.3, -2.2,
     -2.8, -0.9, -1.1, -1.5,  0.0, -0.8, -0.5, -1.0],
    [0.1,  2.2, 1.5, 0.4, 1.1, 0.4, 0.9, 0.5, 0.2, 0.2, -0.2, -0.2, 0.1, -1.4,
     -1.8, -2.1, -1.0, -2.8, -1.9, -3.3],
    [-0.5,  1.5,  2.5, 0.1, 0.6, -1.1, 0.5, 0.0, -0.1, 0.0, -0.3, -0.2, 0.1,
     -0.6, -0.6, -1.3, 0.0, -2.2, -1.9, -3.5],
    [-3.1,  0.4,  0.1,  7.6, 0.3, -1.6, -0.9, -0.7, -0.5, -0.2, -1.1, -0.9,
     -0.6, -2.4, -2.6, -2.3, -1.8, -3.8, -3.1, -5.0],
    [0.5,  1.1,  0.6,  0.3,  2.4, 0.5, -0.3, -0.3, 0.0, -0.2, -0.8, -0.6,
     -0.4, -0.7, -0.8, -1.2, 0.1, -2.3, -2.2, -3.6],
    [-2.0,  0.4, -1.1, -1.6,  0.5,  6.6, 0.4, 0.1, -0.8, -1.0, -1.4, -1.0,
     -1.1, -3.5, -4.5, -4.4, -3.3, -5.2, -4.0, -4.0],
    [-1.8,  0.9,  0.5, -0.9, -0.3,  0.4,  3.8, 2.2, 0.9, 0.7, 1.2, 0.3, 0.8,
     -2.2, -2.8, -3.0, -2.2, -3.1, -1.4, -3.6],
    [-3.2,  0.5,  0.0, -0.7, -0.3,  0.1,  2.2,  4.7, 2.7, 0.9, 0.4, -0.3, 0.5,
     -3.0, -3.8, -4.0, -2.9, -4.5, -2.8, -5.2],
    [-3.0,  0.2, -0.1, -0.5,  0.0, -0.8,  0.9,  2.7,  3.6, 1.7, 0.4, 0.4, 1.2,
     -2.0, -2.7, -2.8, -1.9, -3.9, -2.7, -4.3],
    [-2.4,  0.2,  0.0, -0.2, -0.2, -1.0,  0.7,  0.9,  1.7,  2.7, 1.2, 1.5,
     1.5, -1.0, -1.9, -1.6, -1.5, -2.6, -1.7, -2.7],
    [-1.3, -0.2, -0.3, -1.1, -0.8, -1.4,  1.2,  0.4,  0.4,  1.2,  6.0, 0.6,
     0.6, -1.3, -2.2, -1.9, -2.0, -0.1, 2.2, -0.8],
    [-2.2, -0.2, -0.2, -0.9, -0.6, -1.0,  0.3, -0.3,  0.4,  1.5,  0.6,  4.7,
     2.7, -1.7, -2.1, -2.2, -2.0, -3.2, -1.8, -1.6],
    [-2.8, 0.1,  0.1, -0.6, -0.4, -1.1,  0.8,  0.5,  1.2,  1.5,  0.6,  2.7,
     3.2, -1.4, 2.8, -2.1, -1.7, -3.3, -2.1, -3.5],
    [-0.9, -1.4, -0.6, -2.4, -0.7, -3.5, -2.2, -3.0, -2.0, -1.0, -1.3, -1.7,
     -1.4,  4.3, 2.8, 2.8, 1.6, 1.6, -0.2, -1.0],
    [-1.1, -1.8, -0.6, -2.6, -0.8, -4.5, -2.8, -3.8, -2.7, -1.9, -2.2, -2.4,
     -2.1,  2.5,  4.0, 2.8, 3.1, 1.0, -0.7, -1.8],
    [-1.5, -2.1, -1.3, -2.3, -1.2, -4.4, -3.0, -4.0, -2.8, -1.6, -1.9, -2.2,
     -2.1,  2.8,  2.8,  4.0, 1.8, 2.0, 0.0, -0.7],
    [0.0, -1.0,  0.0, -1.8,  0.1, -3.3, -2.2, -2.9, -1.9, -1.5, -2.0, -2.0,
     -1.7,  1.6,  3.1,  1.8,  3.4, 0.1, -1.1, -2.6],
    [-0.8, -2.8, -2.2, -3.8, -2.3, -5.2, -3.1, -4.5, -3.9, -2.6, -0.1, -3.2,
     -3.3,  1.6,  1.0,  2.0,  0.1,  7.0, 5.1, 3.6],
    [-0.5, -1.9, -1.9, -3.1, -2.2, -4.0, -1.4, -2.8, -2.7, -1.7,  2.2, -1.8,
     -2.1, -0.2, -0.7,  0.0, -1.1,  5.1,  7.8, 4.1],
    [-1.0, -3.3, -3.5, -5.0, -3.6, -4.0, -3.6, -5.2, -4.3, -2.7, -0.8, -1.6,
     -3.5, -1.0, -1.8, -0.7, -2.6,  3.6,  4.1,  14.2]]


def similarity_set(lstMat, lstAA):
    """
    Creates a sets of similar amino acids for pattern extension.

    Similarity sets are created for all amino acid pairs
    p\ :sub:`ij`, if the score ``s`` in similarity matrix ``lstMat`` is
    s\ :sub:`ij`\ >0.5. Any other amino acid ``x`` is added to the set
    if s\ :sub:`ix`\ > s\ :sub:`ij` **and**
    s\ :sub:`jx`\ > s\ :sub:`ij`\ .

    Args:
        lstMat (list): Matrix of amino acid similarities.
        lstAA (list): List with mapping of amino acids for ``lstMat``

    Returns:
        dict: Dictionary with lists of similar amino acids. Keys are
        the amino acid pairs.
    """
    dctSimSets = {}
    for i in xrange(len(lstAA)):
        for j in xrange(i+1):
            if i != j and lstMat[i][j] > 0.5:
                idx1 = max(i, j)
                idx2 = min(i, j)

                fltCutoff = lstMat[idx1][idx2]
                lstSimilar = []

                for aa in xrange(len(lstMat[idx1])):
                    if lstMat[idx1][aa] > fltCutoff and idx1 != aa:
                        lstSimilar.append(aa)

                setSimilar = set(lstSimilar)
                lstSimilar = []

                for aa in xrange(len(lstMat[idx2])):
                    if lstMat[idx2][aa] >= fltCutoff and idx2 != aa:
                        lstSimilar.append(aa)
                setIntersection = setSimilar.intersection(lstSimilar)
                lstSet = "".join([lstAA[s] for s in setIntersection])
                if lstSet:
                    strAAs = min(lstAA[i], lstAA[j]) + max(lstAA[i], lstAA[j])
                    dctSimSets[strAAs] = lstSet
    return dctSimSets


def extend():
    """
    Extends all patterns using similarity sets for semi-conserved
    amino acid positions.

    All patterns are fetched and the pattern positions are extended
    using Gonnets PAM250 similarity matrix. The extended patterns are
    saved to the working database.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.execute("TRUNCATE TABLE patterns_extended")
    cur.execute("SELECT node_id, pattern_ps, min_len, max_len "
                "FROM patterns")
    log.info("Creating extended patterns...")
    lstPatterns = []

    dctSimSets = similarity_set(lstPam250, lstAA)

    for node_id, strPat, min_len, max_len in cur.fetchall():
        lstPat = []
        for item in strPat.split("-"):
            if "[" in item:
                strGrp = "".join(sorted(item[1:-1]))
                lstPairs = get_pairs(strGrp)
                setExt = set(strGrp)
                for strPair in lstPairs:
                    if strPair in dctSimSets:
                        setExt = setExt | set(dctSimSets[strPair])
                lstPat.append("[{}]".format("".join(sorted(setExt))))
            else:
                lstPat.append(item)
        strPat = "-".join(lstPat)
        comp = bl.complexity(strPat)
        lstPatterns.append(
            (node_id, strPat, bl.ps2regex("-".join(lstPat)),
             comp, min_len, max_len))

    # Save extended_patterns
    lstChunks = [lstPatterns[i:i+500]
                 for i in xrange(0, len(lstPatterns), 500)]

    for chunk in lstChunks:
        cur.executemany("INSERT INTO patterns_extended (node_id, pattern_ps, "
                        "pattern_re, complexity, min_len, max_len) VALUES "
                        "(%s,%s,%s,%s,%s,%s)", chunk)
    log.info("Done...")


def get_pairs(strGrp):
    """
    Returns all character combinations of a given string.

    Args:
        strGroup (str): String with characters.

    Raises:
        list: A list with all character combinations.
    """
    lstPairs = []
    for i in strGrp:
        for j in strGrp:
            if (min(i, j) + max(i, j) not in lstPairs and i != j):
                lstPairs.append(min(i, j) + max(i, j))
    return lstPairs


def generate(intThreads=1, useSGE=False):
    """
    Coordinates pattern generation using SGE or multithreading.

    Args:
        intThreads (int): Number of threads for desktop mode
            (Default: 1).
        useSGE (boolean): ``True`` if patterns should be generated on
            SGE computer cluster, ``False`` for desktop mode.
            (Default ``False``)
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    log.info("Preparing pattern generation...")

    cur.execute("TRUNCATE TABLE patterns")
    cur.execute("SHOW INDEX FROM patterns WHERE Column_name=%s", ("node_id",))
    tplRes = cur.fetchall()
    if len(tplRes) > 0:
        cur.execute("ALTER TABLE patterns DROP INDEX {}".format(tplRes[0][2]))

    # Cleanup old files
    for f in glob.glob("{}/*.*".format(SEQ_DIR)):
        os.remove(f)

    if useSGE:
        cur.execute("SELECT COUNT(cluster_id) FROM clusters")
        intClusters = cur.fetchone()[0]
        log.info("Sending {} clusters to the SGE...".format(intClusters))

        os.system(
            "ssh node1 \""
            "(export SGE_ROOT=/opt/sge; cd /home/jke/BrEPSdb;"
            "`which qsub` -t 1-{} "
            "{}/sge_pattern.bash)\"".format(intClusters, os.getcwd())
            )

        # Check if there are temporary helper files to wait for all jobs
        while len(glob.glob("pattern/*.rdy")) < intClusters:
            time.sleep(60)
        for f in glob.glob("pattern/*.rdy"):
            os.remove(f)

    else:
        cur.execute("SELECT cluster_id FROM clusters")
        lstClusters = [x[0] for x in cur.fetchall()]
        log.info("Start pattern generation for {} clusters...".format(
            len(lstClusters))
            )

        # Preparing the job list. For every job entry we need the table
        # with sequence data and a list of cluster ids.
        lstJobs = [[x] for x in lstClusters]

        pool = mp.Pool(processes=intThreads)
        pool.map(process, lstJobs)

    log.info("Pattern generation done!")
    cur.execute("ALTER TABLE patterns ADD INDEX (node_id)")


def process(lstClusters):
    """
    Calls the multiple sequence alignments for cluster nodes.

    Reads the nodes created at clustering from the database and creates
    a fasta file ``<cluster_id>_<node_id>.fsa`` containing the
    associated sequences. Multiple sequence alignment (MSA) will be
    perormed by ``clustalo``. Pattern will be generated with MSA output
    file.

    Args:
        lstClusters (list): List of cluster ids to process.

    Raises:
        IOError: If no MSA output was generated.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.execute("SELECT cluster_id, nodes FROM clusters "
                "WHERE cluster_id IN ({})".format(
                    ",".join(["%s"]*len(lstClusters))),
                lstClusters)
    tplClusters = cur.fetchall()

    numSingles = 0
    numPatterns = 0
    numSkips = 0
    allNumPatterns = 0
    allNumSkips = 0

    cur.execute("SELECT seq_id, ecs, sequence FROM sequences")
    dctSeqs = {x[0]: [x[1].split(";"), x[2]] for x in cur.fetchall()}

    lstPatterns = []
    for tplCluster in tplClusters:
        cluster_id, numNodes = tplCluster

        # We don't use single node clusters
        if numNodes == 1:
            numSingles += 1
            continue

        # We get all nodes which are no leaf nodes
        cur.execute("SELECT node_id,parent_id,left_id,right_id "
                    "FROM nodes WHERE cluster_id=%s "
                    "AND seq_id IS NULL ORDER BY node_id DESC",
                    (cluster_id,))
        tplNodes = cur.fetchall()

        dctEcs = {}

        for node_id, parent_id, left_id, right_id in tplNodes:
            cur.execute("SELECT seq_id FROM nodes_seqs_mapper "
                        "WHERE node_id=%s", (node_id,))
            lstSeqIds = [s[0] for s in cur.fetchall()]
            lstEcs = list({e for s in lstSeqIds
                           for e in dctSeqs[s][0]})
            dctEcs[node_id] = lstEcs

            # Only start MSA if we've got less ECs than before or there
            # is no parent saved (for root nodes or if nodes yielded no
            # pattern)

            if (parent_id not in dctEcs or
                    len(dctEcs[node_id]) < len(dctEcs[parent_id])):

                msa_file = "{}/{}_{}.fsa".format(SEQ_DIR, cluster_id, node_id)
                fh = open(msa_file, "wb")

                lstLines = [">{}\n{}\n".format(s, dctSeqs[s][1])
                            for s in lstSeqIds]

                fh.write("".join(lstLines))
                fh.close()

                os.system("{0}/clustalo -i {1} -o {1}.msa "
                          # "--guidetree-out={1}.tree "
                          "--force "
                          "--outfmt=clu ".format(CLUSTAL_DIR, msa_file))

                if not os.path.isfile("{}.msa".format(msa_file)):
                    log.error("MSA output (%s) not found!", msa_file)
                    raise IOError("MSA output  not found!")

                newPattern = create_pattern("{}.msa".format(msa_file), False)

                os.remove(msa_file)

                # If we got no pattern, remove ECs from list, because
                # we want to get a pattern for every EC combination
                if not newPattern:
                    del dctEcs[node_id]
                else:
                    lstPatterns.append(newPattern)
                    numPatterns += 1
                    # add pattern to lstPatterns
                    if len(lstPatterns) >= 500:
                        log.debug("Writing %s patterns to db!",
                                  len(lstPatterns))
                        cur.executemany(
                            "INSERT INTO complexity (node_id,pattern_ps,"
                            "pattern_re,specificity,min_len,max_len)"
                            "VALUES (%s,%s,%s,%s,%s,%s)", lstPatterns)
                        lstPatterns = []
            else:
                numSkips += 1

        allNumPatterns += numPatterns
        allNumSkips += numSkips

        numPatterns = 0
        numSkips = 0

    cur.executemany(
        "INSERT INTO patterns (node_id,pattern_ps,"
        "pattern_re,complexity,min_len,max_len) "
        "VALUES (%s,%s,%s,%s,%s,%s)", lstPatterns)

    cur.close()
    cur.close()


def create_pattern(strFilename, blnRemove=True):
    """
    Processes the MSA result to create the pattern.

    Reads the MSA result files and generates patterns based on the
    aligned sequences. A pattern needs to have a minimum length of 8
    positions from the first conserved or semi-conserved position to the
    last conserved or semi-conserved position.

    The consensus positions are intepreted  with ``\*`` (Asterisk) as
    conserved position, ``\:`` (Colon) as semi-conserved position, blank
    and ``.`` (Colon) as non-conserved position.

    Args:
        strFilename (str): Path to MSA file to process.

    Returns:
        tuple: Tuple with pattern data or empty tuple if no pattern
        generated.
    """

    regex_ids = re.compile("([0-9]*)_([0-9]*)\.fsa\.msa")

    cluster_id, node_id = regex_ids.search(strFilename).groups()

    try:
        fh = open(strFilename, "r")
    except IOError, e:
        log.error("Could not open file %s!", strFilename)
        raise e

    lstLines = fh.readlines()

    if blnRemove:
        os.remove(strFilename)

    # Fetch length of string before sequence
    m = re.match("\d* *", lstLines[3])

    intStart = len(m.group(0))
    dctSeqences = {}
    strConservation = ""

    # We want to start from the third line, save lines with a
    # seq_id into a dictionary

    for line in lstLines[3:]:
        if line[:intStart].strip().isdigit():
            seq_id = int(line[:intStart].strip())
            if seq_id not in dctSeqences:
                dctSeqences[seq_id] = ""
            dctSeqences[seq_id] += line[intStart:].strip("\n")
        else:
            strConservation += line[intStart:].strip("\n")
    fh.close()
    m = re.search("[\*:]+[\*:\. ]{8,}[\*:]+", strConservation)

    # We stop the script here if we don't get a pattern

    if m is None:
        return ()

    numPos = strConservation.index(m.group(0))
    strConservation = m.group(0)

    # We want to generate the pattern from the conservation string.
    # We will use these conversion rules:
    #     - * = one letter code (100% conserved)
    #     - : = one letter code of all sequences in brackets []
    #     - . or blank = x or - if there is a gap
    # We store every position in a list or add it to the latest
    # list item if it's the same symbol. We also combine x and -.

    lstPattern = []
    for strChar in strConservation:
        if strChar == "*":
            curChar = dctSeqences.values()[0][numPos]
        elif strChar == ":":
            seen = set()
            curChar = "[%s]" % "".join(
                sorted([x[numPos] for x in dctSeqences.values()
                        if not (x[numPos] in seen or seen.add(x[numPos]))])
                )
        elif "-" in [x[numPos] for x in dctSeqences.values()]:
            curChar = "-"
        else:
            curChar = "x"

        if (len(lstPattern) > 0 and (curChar in ["-", "x"] and
            lstPattern[-1][-1] in ["-", "x"] or
                curChar == lstPattern[-1][-1])):
            lstPattern[-1] += curChar
        else:
            lstPattern.append(curChar)

        numPos += 1

    # Finally we compress the pattern list items to generate the
    # final pattern. Here we use - to generate (min,max) ranges of
    # any aminoacid with gaps and (num) without gaps. Entries in
    # brackets will never be merged together.

    min_len = 0
    max_len = 0

    for i in xrange(len(lstPattern)):
        if not lstPattern[i].startswith("["):
            if "-" in lstPattern[i]:
                mx = len(lstPattern[i])
                mn = mx-lstPattern[i].count("-")
                strChar = "x(%s,%s)" % (mn, mx)
                min_len += mn
                max_len += mx
            elif len(lstPattern[i]) == 1:
                strChar = lstPattern[i]
                min_len += 1
                max_len += 1
            else:
                strChar = "%s(%s)" % (lstPattern[i][0], len(lstPattern[i]))
                min_len += len(lstPattern[i])
                max_len += len(lstPattern[i])
            lstPattern[i] = strChar
        else:
            # Sort characters in brackets
            strChars = "".join(sorted(lstPattern[i][1:-1]))
            lstPattern[i] = "[{}]".format(strChars)
            min_len += 1
            max_len += 1

    pattern_ps = "-".join(lstPattern)
    pattern_re = bl.ps2regex(pattern_ps)

    comp = bl.complexity(pattern_ps)

    return (node_id, pattern_ps, pattern_re, comp, min_len, max_len)

if __name__ == "__main__":
    # Simple main program for use on SGE cluster nodes. The shell script
    # basically has to call theis file with one argument: The cluster id
    # to process.

    intCluster = int(sys.argv[1])
    process([intCluster])
