#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
``verification.py``
-------------------
This module contains the verification process for all generate
patterns.
"""
import re
import os
import sys
import json
import time
import logging
import signal
import itertools
import logging.config
import multiprocessing as mp

import brepslib as bl

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


# Timeout exception for the verification timeout.
class TimeoutExc(Exception):
    pass


def timeout(intSecs, *args):
    """
    Timeout decorator for the pattern verification.

    Complex regular expression search may run for a long time causing
    the verification to get stuck. The decorator adds a signal that
    kills a verification that runs longer than ``intSecs`` seconds.

    Args:
        intSecs (integer): Seconds after which the timeout is fired.
    """
    def decorate(f):
        def handler(signum, frame):
            raise TimeoutExc("Timeout!")

        def new_f(*args):
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(intSecs)
            return f(*args)
            signal.alarm(0)

        new_f.__name__ = f.__name__
        return new_f
    return decorate


def verificator(strTable="patterns", intThreads=1, useSGE=False):
    """
    Coordinates pattern verification using SGE or multithreading.

    Args:
        strTable (string): Table with patterns to verify.
            (Default ``patterns``).
        intThreads (int): Number of threads for desktop mode.
            (Default 1).
        useSGE (boolean): ``True`` if patterns should be generated on
            SGE computer cluster, ``False`` for desktop mode.
            (Default ``False```).
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    log.info("Starting pattern verification of table %s...", strTable)

    if useSGE:
        cur.execute("SELECT COUNT(pat_id) FROM {}".format(strTable))
        intPatterns = cur.fetchone()[0]
        # Submit a job to the Sun Grid Engine
        os.system(
            "ssh node1 \""
            "(export SGE_ROOT=/opt/sge; cd /home/jke/BrEPSdb;"
            "`which qsub` -t 1-{} "
            "{}/sge_verification.bash {})\"".format(
                intPatterns,
                os.getcwd(),
                strTable
                )
            )
        cur.execute("SELECT COUNT(pat_id) FROM {} "
                    "WHERE tp IS NULL".format(strTable))
        while cur.fetchone()[0] > 0:
            time.sleep(60)
            cur.execute("SELECT COUNT(pat_id) FROM {} "
                        "WHERE tp IS NULL".format(strTable))
    else:
        cur.execute("SELECT COUNT(pat_id) FROM {}".format(strTable))
        intPatterns = cur.fetchone()[0]
        pool = mp.Pool(processes=intThreads)
        lstJobs = [(strTable, _) for _ in xrange(1, intPatterns+1)]
        pool.map(verify_pattern, lstJobs)
    cur.close()
    log.info("Pattern verification done!")


@timeout(1)
def re_match(reSub, rePat, strSeq):
    if ((bool(re.search(reSub, strSeq)) and
         bool(re.search(rePat, strSeq)))):
        return True
    else:
        return False


def verify_pattern(tplData):
    """
    Verifies a pattern against all sequences.

    Checks every pattern against all sequences in the ``sequences``
    table. Calculates true positives (tp), false positives (fp) and
    false negatives (fn) to rank the patterns. The correstonding pattern
    gets updated values for tp, fp, and fn.

    Args:
        tplData (tuple): Tuple containing two values. Table name of the
            pattern and pattern id to verify.
    """
    strTable, intPatId = tplData
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])
    cur.execute("SELECT pat_id, node_id, pattern_re, pattern_ps, min_len "
                "FROM {} WHERE pat_id={}".format(strTable, intPatId))
    lstPatterns = cur.fetchall()

    # Get all sequences and ECs as pipe separated list
    cur.execute("SELECT seq_id, ecs, sequence, seq_len, enzyme FROM sequences")
    dctSeqs = {s[0]: s[1:] for s in cur.fetchall()}

    # Prepare regex for all unique EC number combinations of all
    # sequences in the database, using the EC string as the key
    dctRegex = {
        dctSeqs[k][0]: re.compile(dctSeqs[k][0].replace(
            ".", "\.").replace("-", "\d*").replace(";", "|"))
        for k in dctSeqs if dctSeqs[k][3] == 1
        }
    lstUpdates = []

    for pat_id, node_id, pattern_re, pattern_ps, min_len in lstPatterns:

        tp = 0
        fp = 0
        fn = 0
        na = 0
        non_enzyme_hits = 0
        # Get ECs from node_id using nodes_seqs_mapper
        cur.execute("SELECT seq_id FROM nodes_seqs_mapper WHERE node_id=%s",
                    (node_id,))

        lstPatEcs = list({e for s in cur.fetchall()
                         for e in dctSeqs[s[0]][0].split(";")})

        strPatEcs = "|".join(lstPatEcs)

        # Check if we want to use reversed pattern/sequence
        lstPattern = pattern_ps.split("-")
        intNGaps = len([i for i in itertools.takewhile(lambda x: "x" in x,
                       lstPattern[1::2])])
        intCGaps = len([i for i in itertools.takewhile(lambda x: "x" in x,
                       lstPattern[::-1][1::2])])
        if intNGaps > 3 and intNGaps > intCGaps:
            rev = True
            pattern_ps = "-".join(lstPattern[::-1])
            pattern_re = bl.ps2regex(pattern_ps)
        else:
            rev = False

        rePat = re.compile(pattern_re)

        reSub = re.compile(bl.ps2regex(bl.subpattern(pattern_ps)))
        reECs = re.compile(
            strPatEcs.replace(".", "\.").replace("-", "[\d*|-]"))
        t0 = time.time()
        seqs_tested = 0
        # Iter all sequences
        for seq_id in dctSeqs:
            seq_ecs, seq, seq_len, seq_enzyme = dctSeqs[seq_id]
            pattern_match = False

            if rev:
                seq = seq[::-1]

            # Check if the pattern hits, but only if pattern legth
            # >= sequence length!
            if (min_len <= seq_len):
                seqs_tested += 1

            if min_len <= seq_len:
                try:
                    pattern_match = re_match(reSub, rePat, seq)
                except TimeoutExc:
                    pattern_match = False

            # Continue loop if we got a non-enzyme, add fp if pattern
            # hits.
            if pattern_match and seq_enzyme == 0:
                fp += 1
                non_enzyme_hits += 1
                continue
            elif not pattern_match and seq_enzyme == 0:
                continue

            for_comp = bool(reECs.search(seq_ecs))

            if pattern_match and for_comp:
                # Pattern does match and ecs match
                tp += 1
            elif not pattern_match and for_comp:
                # Pattern dosn't match but ecs match
                fn += 1
            elif (pattern_match and not bool(
                    dctRegex[seq_ecs].search(strPatEcs))):
                # Pattern does match but ecs doesn't
                fp += 1
            else:
                # In all cases the hit is neutral
                na += 1
            t1 = time.time()
            if t1-t0 > 300:
                log.debug("Pattern %s: verification stopped after %.2fs!",
                          pat_id, t1-t0)
                cur.execute("UPDATE {} SET tp=0 "
                            "WHERE pat_id=%s".format(strTable), (pat_id,))
                break

        lstUpdates.append((tp, fp, fn, na, pat_id))
        if len(lstUpdates) >= 500:
            log.debug("Updating %s datasets!", len(lstUpdates))
            cur.executemany("UPDATE {} SET tp=%s, fp=%s, fn=%s, na=%s "
                            "WHERE pat_id=%s".format(strTable), lstUpdates)
            lstUpdates = []

    cur.executemany("UPDATE {} SET tp=%s, fp=%s, fn=%s, na=%s "
                    "WHERE pat_id=%s".format(strTable), lstUpdates)
    cur.close()

if __name__ == "__main__":
    # Simple main program for use on SGE cluster nodes. The shell script
    # basically has to call theis file with two arguments:
    # The pattern id to verify and the sequence table to use.
    intPatId = int(sys.argv[1])
    strTable = sys.argv[2]
    verify_pattern((strTable, intPatId))
