#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``blast.py``
------------
This module prepares FASTA files and executes ``blastp`` for the
BLAST alignment.

The ``blast`` folder is associated to this module.
"""
import os
import gc
import sys
import json
import glob
import time
import logging
import logging.config

import brepslib as bl

strPath = os.path.abspath(os.path.dirname(sys.modules[__name__].__file__))

logconfig = json.load(open("{}/config/logging.json".format(strPath), "rb"))
logging.config.dictConfig(logconfig)

dctCfg = json.load(open("{}/config/breps.json".format(strPath), "rb"))

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

BLAST_DIR = "{}/blast-current/bin".format(strPath)
BLAST_EXE = "{}/blastp".format(BLAST_DIR)

SEQ_DIR = "{}/blast".format(strPath)
BLAST_DB = "{}/blast_db.fsa".format(SEQ_DIR)


def prepare(intFiles=1):
    """
    Prepares all files for the blast alignment.

    All sequences will be written to a database file in FASTA format.
    In parallel the sequences will be written into ``<intFiles>``
    separate files for BLASTing on multiple SGE cluster nodes or
    threads. The database will get masked by the BLAST ``segmasker``
    to mask repetetive regions in sequences. Finally the database file
    will be created.

    Args:
        intFiles (int): Number of different sequence files to
            generate for BLASTing (Default 1).

    Returns:
        list: list of created FASTA files.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    log.info("Preparing data for BLAST...")

    for f in glob.glob("{}/*.*".format(SEQ_DIR)):
        os.remove(f)

    fh_db = open("{}/blast_db.fsa".format(SEQ_DIR), "wb")

    # We just select non redundant sequences from database and
    # distribute them between <intFiles> FASTA files.
    cur.execute("SELECT seq_id,sequence FROM sequences WHERE enzyme=1 "
                "GROUP BY sequence ORDER BY seq_id ")
    lstSeqs = cur.fetchall()

    lstPerFile = [len(lstSeqs) / intFiles for _ in xrange(intFiles)]
    for i in xrange(len(lstSeqs) % intFiles):
        lstPerFile[i] += 1
    intSeqsDone = 0
    lstFiles = []
    intFile = 1
    for intSeqs in lstPerFile:
        fh = open("{}/sequences_{}.fsa".format(SEQ_DIR, intFile), "wb")
        lstFiles.append("{}/sequences_{}.fsa".format(SEQ_DIR, intFile))
        for seq in lstSeqs[intSeqsDone:intSeqsDone+intSeqs]:
            fh_db.write(">{0}\n{1}\n".format(*seq))
            fh.write(">{0}\n{1}\n".format(*seq))
            intSeqsDone += 1
        intFile += 1

    fh_db.close()
    cur.close()

    # We mask the repetitive sequences using the segmaster tool,
    # creating a binary ASN file. Afterwards the database is created
    # using the makeblastdb command.
    log.debug("Starting segmasker...")
    seg_cmd = ("{0}/segmasker -in {1}/blast_db.fsa -infmt fasta "
               "-outfmt maskinfo_asn1_bin "
               "-out {1}/blast_db.segmask.asnb").format(BLAST_DIR, SEQ_DIR)
    os.system(seg_cmd)

    log.debug("Create blastdb...")

    makedb_cmd = (
        "{0}/makeblastdb -in {1}/blast_db.fsa -input_type fasta "
        "-dbtype prot -mask_data {1}/blast_db.segmask.asnb "
        "-logfile {1}/blast_db.log").format(BLAST_DIR, SEQ_DIR)
    os.system(makedb_cmd)
    log.info("Done BLAST preparations!")
    return lstFiles


def blast(lstFiles, useSGE=False):
    """
    Executes the BLAST either on the SGE or on multiple threads.

    All sequences will be written to a database file in FASTA format.
    In parallel the sequences will be written into ``<intFiles>``
    separate files for BLASTing on multiple SGE cluster nodes or
    threads. The database will get masked by the BLAST ``segmasker``
    to mask repetetive regions in sequences. Finally the database file
    will be created. SGE jobs are created using ``sge_blast.bash``.

    Args:
        lstFiles (list): List of all FASTA files for the BLAST.
        useSGE (boolean): Whether or not to deploy BLAST on the SGE.

    Returns:
        list: list of created BLAST result files.
    """
    log.info("Starting BLAST...")
    intFiles = len(lstFiles)

    if useSGE:
        os.system(
            "ssh node1 \""
            "(export SGE_ROOT=/opt/sge; cd /home/jke/BrEPSdb;"
            "`which qsub` -t 1-{} "
            "{}/sge_blast.bash)\"".format(intFiles, os.getcwd()))

        # Check if we got all result files we need to proceed...
        while len(glob.glob("{}/*.res".format(SEQ_DIR))) != intFiles:
            time.sleep(60)
    else:
        for strFile in lstFiles:
            f = "{}".format(strFile)
            os.system("{0}/blastp -db {1}/blast_db.fsa -query {2} "
                      "-outfmt \"6 qseqid sseqid evalue nident\" "
                      "-out {2}.res "
                      "-max_target_seqs 100000 "
                      "-num_threads 4 "
                      # "-num_alignments 1 "
                      "-evalue 0.0001".format(BLAST_DIR, SEQ_DIR, f)
                      )
            os.remove(f)

    log.info("BLAST done...")

    return glob.glob("{}/*.res".format(SEQ_DIR))


def save(lstFiles=None):
    """
    Reads the alignment results and saves the BLAST alignments to the
    database.

    Parses the ``<lstFiles>`` and saves alignments to the database.
    Results for seq1/seq2 and seq2/seq1 will be merged into one result
    to produce a symmetrical matrix and saving just one side of the
    matrix with seq1 = min(seq1,seq2), seq2 = max(seq1,seq2) and using
    the lower E-value. E-values of 0.0 will be 1 for self alignments and
    1e-181 for all other alignments.

    Args:
        lstFiles (list): List of files with BLAST results.

    Returns:
        boolean: ``True`` if successfull.

    Raises:
        IOError: If one file of ``<lstFiles>`` is not found.
    """
    cur = bl.Dbase("mysql", dctCfg["working_db"], dctCfg["db_host"],
                   dctCfg["db_user"], dctCfg["db_pass"])

    # We need to prepare the database, drop the content and take care
    # of the indices. We want to index the sequence ids after filling
    # the database to speed up the insertion of data.

    log.info("Preparing saving of BLAST data...")

    cur.execute("TRUNCATE TABLE alignments")
    cur.execute("SHOW INDEX FROM alignments WHERE Column_name=%s",
                ("seq_id1",))
    if len(cur.fetchall()) > 0:
        cur.execute("ALTER TABLE alignments DROP INDEX seq_id1")
    cur.execute("SHOW INDEX FROM alignments WHERE Column_name=%s",
                ("seq_id2",))
    if len(cur.fetchall()) > 0:
        cur.execute("ALTER TABLE alignments DROP INDEX seq_id2")

    dctResults = {}
    lstResults = []
    n = 0
    log.info("Start parsing of {} result files...".format(len(lstFiles)))
    for strFile in lstFiles:
        log.debug("Start parsing of alignments in {}...".format(strFile))
        try:
            fh = file(strFile, "r")
        except IOError:
            log.error("File %s not found!", strFile)
            exit(0)
        for line in fh:
            if n % 10000000 == 0:
                log.debug("Starting line {:>8}".format(n))
            n += 1

            # Every row in the results file is one alignment in format:
            # query[Tab]subject[Tab]evalue

            qry, sbj, eva, ident = line.split("\t")
            qry = int(qry)
            sbj = int(sbj)
            eva = float(eva)

            # We move the result from dict to list if we had both
            # orientations. This saves iterations in the final list
            # preparation for executemany

            seq_id1 = min(qry, sbj)
            seq_id2 = max(qry, sbj)

            if qry == sbj:
                eva = 1

            if seq_id1 not in dctResults:
                dctResults[seq_id1] = {"seq_id2": eva}
            elif seq_id2 not in dctResults[seq_id1]:
                dctResults[seq_id1][seq_id2] = eva
            else:
                dctResults[seq_id1][seq_id2] = max(
                    eva, dctResults[seq_id1][seq_id2])
                lstResults.append(
                    (seq_id1, seq_id2, max(eva, dctResults[seq_id1][seq_id2]))
                    )
                del dctResults[seq_id1][seq_id2]

            if len(lstResults) > 25000:
                cur.executemany("INSERT INTO alignments "
                                "(seq_id1,seq_id2,evalue) "
                                "VALUES (%s,%s,%s)", lstResults)
                del lstResults
                lstResults = []

    cur.executemany("INSERT INTO alignments (seq_id1,seq_id2,evalue) "
                    "VALUES (%s,%s,%s)", lstResults)

    del lstResults
    del dctResults

    gc.collect()
    log.info("Start indexing alignment table...")
    cur.execute("ALTER TABLE alignments ADD INDEX (seq_id1), "
                "ADD INDEX (seq_id2)")

    log.info("Saved BLAST data!")

    fh.close()
    cur.close()

if __name__ == "__main__":
    pass
